/* #define SERV_DEBUG 1 */

/* ----------------------------------------------------------------------
   The SINUS Firewall -- a TCP/IP packet filter for Linux

   Written within the SINUS project at the University of Zurich,
   SWITCH, Telekurs Payserv AG, ETH Zurich.
   originally based on the sf Firewall Software (C) 1996 by Robert
   Muchsel and Roland Schmid.

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

   SINUS Firewall resources:
   SINUS Homepage: http://www.ifi.unizh.ch/ikm/SINUS/
   Firewall Homepage: http://www.ifi.unizh.ch/ikm/SINUS/firewall.html
   Frequently asked questions: http://www.ifi.unizh.ch/ikm/SINUS/sf_faq.html
   Mailing list for comments, questions, bug reports: firewall@ifi.unizh.ch
   ----------------------------------------------------------------------  */



/*
 * sf Firewall Server - Communication with Configuration Client
 *
 */

#include <stdio.h>
#include <fcntl.h>
#include <unistd.h>
#include <string.h>
#include <errno.h>
#include <limits.h>
#include <stdlib.h>
#include <time.h>
#include <syscall.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <linux/kernel.h>
#include <asm/param.h>

int sysinfo(struct sysinfo *info);

#include "sf_custom.h"
#include "sf_control.h"
#include "sf_global.h"
#include "sf_config.h"
#include "sf_daemon.h" 
#include "sf_relay.h"
#include "sf_server.h"
#include "sf_md5.h"

/*
 * Public variables
 */

int sockserv = 0;		/* socket for accepting connections */
int server_enabled = 0;         /* remote configuration enabled */
int conn_established = 0;	/* number of established client connections */
struct sockaddr_in servaddr;	/* server address */
int cliaddrlen;			/* length of client address */
int lastclient = 0;		/* round robin read from clients */
int reestablish_timeout = 0;    /* flag for event queue */

int sockconn[MAX_CLIENTS];	/* socket handles */
int conn_state[MAX_CLIENTS];	/* state of connection */
int conn_child_pid[MAX_CLIENTS];	/* child is using connection */
FILE *conn_fds[MAX_CLIENTS];	/* file descriptors for WRITE_FILE */
FILE *passwd;						/* firewall passwd */
struct sockaddr_in cliaddr[MAX_CLIENTS];	/* client addresses */
MD5_CTX context;			/* message digest context */
unsigned char digest[16]; /* message digest itself */
struct auth_data passwdbuf;				/* passwd entry */
char conn_keys[MDKEY_SIZE][MAX_CLIENTS];	/* keys to be updated in passwd file */
char conn_logins[LOGIN_MAXLEN+1][MAX_CLIENTS];	/* logins to be updated in passwd file */

/*
 * Static variables and other local stuff
 */

static char buffer[BUFFER_SIZE];	/* static read buffer */
static char *connbuf[MAX_CLIENTS];	/* dynamic buffers */
static int  connbufsize[MAX_CLIENTS];	/* size of connbuf */
static char login[LOGIN_MAXLEN+1];			/* login for authentication */
static char key[MDKEY_SIZE];	/* MD key buffer for authentication*/

#ifdef SERV_DEBUG
static void debug_entry(int i, char *msg, int status)
{ char tmp[255];                /* string for log message construction */
  char stmp[20];                /* string for IP address to string conversion */

  strcpy(stmp, inet_ntoa(cliaddr[i].sin_addr));
  if (status)
    sprintf(tmp, "DEBUG: Client conn. %i to %s:%hu\n"
                 "       %s\n"
                 "       conn_state[%i] == %x, conn_child_pid[%i] == %i",
            i, stmp, ntohs(cliaddr[i].sin_port), 
            msg, 
            i, conn_state[i], i, conn_child_pid[i]);
  else
    sprintf(tmp, "DEBUG: Client conn. %i to %s:%hu\n"
                 "       %s",
            i, stmp, ntohs(cliaddr[i].sin_port), msg);
  log_entry(tmp);
  
}
#endif

/*
 * cleanup after abnormally terminated connection
 */

static void handle_abort(int i)
{
  close(sockconn[i]);
  if ((conn_state[i] & STATE_WAITDATA) == STATE_WAITDATA) 
    free(connbuf[i]);
  if ((conn_state[i] & STATE_WRITE_FILE) == STATE_WRITE_FILE)
    fclose(conn_fds[i]);
  connbufsize[i] = 0;
  conn_established--;
  conn_state[i] = STATE_NOCONN;
  conn_child_pid[i] = 0;
}

/*
 * Read data from socket, returns number of bytes read, negative value on error
 */

#define READ_ECLOSE	-1	/* connection has been closed by client */
#define READ_EFATAL	-2	/* any read error that should not occur,
				   will be handled like aborted connection */

static int read_data(int i)
{
  int retval;
  char tmp[255];		/* string for log message construction */
  char stmp[20];		/* string for IP address to string conversion */

  if ((retval = read(sockconn[i], buffer, sizeof(buffer))) < 0) {

    /* read terminated by signal -> return 0 bytes read */
    if (errno == EINTR)
      return 0;

    if (errno != EPIPE) {

      /* handle any other error like connection termination */

      strcpy(stmp, inet_ntoa(cliaddr[i].sin_addr));
      sprintf(tmp, "Remote config conn. aborted (read error): %s:%hu",
              stmp, ntohs(cliaddr[i].sin_port));
      log_entry(tmp);
  
      handle_abort(i);            /* close sockconn and clean up */

      return READ_EFATAL;
    }
    else
      retval = 0; /* According to the documentation only write() should
                     return EPIPE, but read() does, too! So we handle
                     EPIPE like EOF. */
  }

  if (retval == 0) {
    /* EOF, i.e. client has closed the connection */

    strcpy(stmp, inet_ntoa(cliaddr[i].sin_addr));
    sprintf(tmp, "Remote config conn. aborted by client: %s:%hu",
	    stmp, ntohs(cliaddr[i].sin_port));
    log_entry(tmp);

    handle_abort(i);            /* close sockconn and clean up */

    return READ_ECLOSE;
  }

  /* read successful: assert(retval>0) */
  return retval;

}

/*
 * Write data to socket
 */

#define WRITE_EPIPE	-1	/* write error EPIPE */
#define WRITE_EFATAL	-2	/* unexpected write error, connection closed */

static int write_data(int i, void *buf, size_t count)
{
  void *bf = buf;
  size_t cnt = count;
  int retval = 0;
  char tmp[255];		/* string for log message construction */
  char stmp[20];		/* string for IP address to string conversion */

  while (cnt > 0) {

    if ((retval = write(sockconn[i], bf, cnt)) <= 0) {

      if (errno == EINTR)
	continue;		/* write interrupted by signal, retry */

      strcpy(stmp, inet_ntoa(cliaddr[i].sin_addr));

      if (errno == EPIPE) {	/* client has closed connection */

	sprintf(tmp, "Remote config conn. aborted by client: %s:%hu",
		stmp, ntohs(cliaddr[i].sin_port));
	log_entry(tmp);

        handle_abort(i);            /* close sockconn and clean up */

	return WRITE_EPIPE;
      }
      else {			/* handle any other error like connection termination */

	sprintf(tmp, "Remote config conn. aborted (write error): %s:%hu",
		stmp, ntohs(cliaddr[i].sin_port));
	log_entry(tmp);

        handle_abort(i);            /* close sockconn and clean up */

	return WRITE_EFATAL;
      }
    }

    cnt -= retval;
    bf = (void *) ((unsigned long) bf + (unsigned long) retval);
  }

  return count;
}

/*
 * ping connections and delete on timeout
 */

void conn_timeout(int dummy1, int dummy2)
{ int i;
  struct serv_data sd;
  char tmp[255];                /* string for log message construction */
  char stmp[20];                /* string for IP address to string conversion */

  for (i=0; i < MAX_CLIENTS; i++) {

    if (!conn_state[i])
      continue;

    if ((conn_state[i] & STATE_CHILD) == STATE_CHILD) {
      /* no timeout while child is alive */
      conn_state[i] &= ~STATE_WAITPING;
      continue;
    }

    if ((conn_state[i] & STATE_WAITPING) == STATE_WAITPING) {
      /* connection timed out */
      strcpy(stmp, inet_ntoa(cliaddr[i].sin_addr));
      sprintf(tmp, "Remote config conn. aborted (timeout): %s:%hu",
              stmp, ntohs(cliaddr[i].sin_port));
      log_entry(tmp);
      handle_abort(i);            /* close sockconn and clean up */
      continue;
    }

    sd.magic = htonl(SERV_MAGIC);
    sd.command = htonl(SERV_PING);
    sd.length = 0;
    if (write_data(i, (void *) &sd, sizeof(sd)) > 0) {
      conn_state[i] |= STATE_WAITPING;
#ifdef SERV_DEBUG
      debug_entry(i, "pinged", 0);
#endif
    }
  } 

  reestablish_timeout = 1;
}

/*
 * Initialize server, perform socket, bind, listen on sockserv
 */

void serv_init()
{
  int n;

  for (n=0; n<MAX_CLIENTS; n++) { /* initialize variables */
    sockconn[n] = 0;
    conn_state[n] = STATE_NOCONN;
    conn_child_pid[n] = 0;
    connbuf[n] = NULL;
    connbufsize[n] = 0;
  }

  if ((sockserv = socket(AF_INET, SOCK_STREAM, 0)) < 0) {
    perror("Cannot create server socket");
    log_entry("Cannot create server socket!");
    log_entry("WARNING: Remote configuration disabled!");
    return;
  }

  n = 1;
  if (setsockopt(sockserv, SOL_SOCKET, SO_REUSEADDR,
		 &n, sizeof(n)) < 0) {
    perror("WARNING: Cannot set SO_REUSEADDR option on server socket");
    log_entry("WARNING: Cannot set SO_REUSEADDR option on server socket!");
  }

  memset(&servaddr, 0, sizeof(servaddr));
  servaddr.sin_family = AF_INET;
  servaddr.sin_addr.s_addr = htonl(INADDR_ANY);		/* use own address */
  servaddr.sin_port = htons(SERVER_PORT);

  if (bind(sockserv, (struct sockaddr *) &servaddr, sizeof(servaddr)) < 0) {
    perror("Cannot bind address to server socket");
    log_entry("Cannot bind address to server socket!");
    log_entry("WARNING: Remote configuration disabled!");
    close(sockserv);
    return;
  }

//#ifdef IPSEC_LEVEL_REQUIRE
//  n = IPSEC_LEVEL_REQUIRE;
//  if (setsockopt(sockserv, SOL_SOCKET, SO_SECURITY_AUTHENTICATION, 
//                 &n, sizeof(n)) < 0)
//#endif
//  {
//    perror("Cannot set socket security on server socket");
//    log_entry("Cannot set socket security on server socket!");
//    log_entry("WARNING: Remote configuration disabled!");
//    close(sockserv);
//    return;
//  }

  if (listen(sockserv, SERVER_BACKLOG) < 0) {
    perror("Cannot listen on server socket");
    log_entry("Cannot listen on server socket!");
    log_entry("WARNING: Remote configuration disabled!");
    close(sockserv);
    return;
  }

  /* timeout check for client connections */
  add_event(conn_timeout, 0, 0, CLIENT_TIMEOUT);

  server_enabled = 1;
}

#ifdef SERV_DEBUG
/*
 * Test function
 */
#define ST 2 /* sleep time between tests */

void testconn(int num)
{ struct serv_data sd;
  struct serv_data *sdp1, *sdp2;

  sd.magic  = htonl(SERV_MAGIC);
  sd.length = 0;
  
  /* send client hello first */
  sd.command = htonl(SERV_CLIENT_HELLO);
  if (write_data(num, (void *) &sd, sizeof(sd)) < 0)
    return; 
  sleep(ST);

  /* illegal magic */
  sd.magic = htonl(0xabc);
  if (write_data(num, (void *) &sd, sizeof(sd)) < 0)
    return;
  sleep(ST);

  /* length > MAX */
  sd.magic = htonl(SERV_MAGIC);
  sd.length = htonl(6000);
  if (write_data(num, (void *) &sd, sizeof(sd)) < 0)
    return;
  sleep(ST);

  /* split packet */
  sdp1 = (struct serv_data *)buffer;
  sdp1->magic = htonl(SERV_MAGIC);
  sdp1->length = 0;
  sdp1->command = htonl(SERV_PING_REPLY);
  if (write_data(num, (void *) buffer, 2) < 0)
    return;
  sleep(1);
  if (write_data(num, (void *) &(buffer[2]), 10) < 0)
    return;
  sleep(ST);

  /* split packet 2 */
  sdp1->length = htonl(100);
  sdp1->command = htonl(99);
  if (write_data(num, (void *) buffer, 32) < 0)
    return;
  sleep(1);
  if (write_data(num, (void *) &(buffer[32]), 80) < 0)
    return;
  sleep(ST);

  /* wrong length */
  if (write_data(num, (void *) buffer, 150) < 0)
    return;
  sleep(ST);

  /* two packets */
  sdp2 = (struct serv_data *) &(buffer[112]);
  sdp2->magic = htonl(SERV_MAGIC);
  sdp2->length = htonl(10);
  sdp2->command = htonl(98);
  if (write_data(num, (void *) buffer, 134) < 0)
    return;
  sleep(ST);

  /* two packets, first split */
  if (write_data(num, (void *) buffer, 50) < 0)
    return;
  sleep(1);
  if (write_data(num, (void *) &(buffer[50]), 84) < 0)
    return;
  sleep(ST);

  /* send big packet */
  memset(buffer, 65, BUFFER_SIZE);
  sdp1->magic = htonl(SERV_MAGIC);
  sdp1->length = htonl(BUFFER_SIZE - sizeof(struct serv_data));
  sdp1->command = htonl(33);
  if (write_data(num, buffer, BUFFER_SIZE) < 0)
    return;
  sleep(ST);

  /* send close */
  sdp1->length = 0;
  sdp1->command = htonl(SERV_CLOSE);
  conn_state[num] |= STATE_CLOSEWAIT;
  if (write_data(num, buffer, 12) < 0)
    return;
  sleep(ST);

}
#endif

/*
 * Initialize client-server communication
 */

void serv_new_connect(int i)
{ struct serv_data sd;

#ifdef SERV_DEBUG
  /* testconn(i); */
#endif

  sd.magic = htonl(SERV_MAGIC);
  sd.command = htonl(SERV_SERVER_HELLO);
  sd.length = 0;

  if (write_data(i, (void *) &sd, sizeof(sd)) < 0)
    return;

  conn_state[i] |= STATE_WAITHELLO;
}

/*
 * Send file to client
 */

static void send_file(int i, char *filename)
{ char lbuf[BUFFER_SIZE]; /* use local buffer ! */
  struct serv_data *sd = (struct serv_data *)lbuf;
  unsigned long sdlength;
  FILE *f;
  char tmp[255];
  char stmp[20];

  sd->magic = htonl(SERV_MAGIC);
  sd->command = htonl(SERV_FILE_DATA);

  // check permission
  if ((strncmp(filename, CONF_DIR, strlen(CONF_DIR)) != 0) && 
      (strncmp(filename, LOG_DIR, strlen(LOG_DIR)) != 0)) {
    strcpy(stmp, inet_ntoa(cliaddr[i].sin_addr));
    sprintf(tmp, "Illegal read attempt of %s from %s:%hu", 
            filename, stmp, ntohs(cliaddr[i].sin_port));
    log_entry(tmp);
    sd->length = 0;
    (void)write_data(i, lbuf, sizeof(struct serv_data));
    return;
  }

  if ((f = fopen(filename, "r")) == NULL) {
    sd->length = 0;
    (void)write_data(i, lbuf, sizeof(struct serv_data));
    return;
  }

  do {
    sdlength = fread(sd->data, 1, BUFFER_SIZE - sizeof(struct serv_data), f);
    sd->length = htonl(sdlength);
    if (write_data(i, lbuf, sizeof(struct serv_data) + sdlength) < 0) {
      fclose(f);
      return;
    }
  } while (sdlength > 0);

  fclose(f);
}

/*
 * Send address array to client
 */

static void send_addr(int i)
{ char lbuf[BUFFER_SIZE]; /* use local buffer ! */
  struct serv_data *sd = (struct serv_data *)lbuf;
  struct sf_address *sa = (struct sf_address *)(sd->data);
  unsigned long sdlength;
  int j;

  sd->magic = htonl(SERV_MAGIC);
  sd->command = htonl(SERV_FILE_DATA);

  sdlength = 4;
  *((int *)(sd->data)) = htonl((unsigned long)sf_addr_free);
  sd->length = htonl(sdlength);
  if (write_data(i, lbuf, sizeof(struct serv_data) + sdlength) < 0) 
    return;

  sdlength = 12;
  sd->length = htonl(sdlength);

  /* sf_addr[0].addr contains number of internalnet addresses in host byte order */
  sa->addr = htonl(sf_addr[0].addr);
  sa->mask = htonl(sf_addr[0].mask);
  sa->port = htons(sf_addr[0].port);
  sa->prend = htons(sf_addr[0].prend);
  if (write_data(i, lbuf, sizeof(struct serv_data) + sdlength) < 0)
    return;

  for (j=1; j<sf_addr_free; j++) {
    sa->addr = sf_addr[j].addr; /* sf_addr contains addresses in network byte order */
    sa->mask = sf_addr[j].mask;
    sa->port = htons(sf_addr[j].port);
    sa->prend = htons(sf_addr[j].prend);
    if (write_data(i, lbuf, sizeof(struct serv_data) + sdlength) < 0)
      return;
  }

  sdlength = 0;
  sd->length = htonl(sdlength);
  write_data(i, lbuf, sizeof(struct serv_data));
  
}

/*
 * Send rules to client
 */

static void send_rules(int i)
{ char lbuf[BUFFER_SIZE]; /* use local buffer ! */
  struct serv_data *sd = (struct serv_data *)lbuf;
  struct sf_fw *r;
  struct sf_fw krule;
  struct sf_fw_trans {
    int fw_line;
    int level_num;
    long timeout;
    int fw_rc;
    long fw_src_idx;
    long fw_src_cnt;
    long fw_dst_idx;
    long fw_dst_cnt;
    unsigned long fw_flags;
    unsigned long ttl;
    unsigned long protocol;
    long fw_rip_idx;
    long fw_rip_cnt;
    unsigned long usage;
    unsigned long bytes;
    unsigned long gbytes;
  };
  struct sf_fw_trans *rt = (struct sf_fw_trans *)(sd->data);
  unsigned long sdlength;
  int fd, result;

  sd->magic = htonl(SERV_MAGIC);
  sd->command = htonl(SERV_FILE_DATA);

  if ((fd = open(sf_devname, O_RDONLY)) < 0) {
    sdlength = 0;
    sd->length = htonl(sdlength);
    write_data(i, lbuf, sizeof(struct serv_data));
    return;
  }

  sdlength = sizeof(struct sf_fw_trans);
  sd->length = htonl(sdlength);

  do {
    result = read(fd, &krule, sizeof(krule));
    if (result < 0) {
      sdlength = 0;
      sd->length = htonl(sdlength);
      write_data(i, lbuf, sizeof(struct serv_data));
      return;
    }
    if (result > 0) {
      r = krule.rule_id.ptr;
      rt->fw_line = htonl(krule.fw_line);
      if (r->level.ptr != NULL)
        rt->level_num = htonl(r->level.ptr->n_level);
      else
        rt->level_num = 0;
      rt->timeout = htonl(r->timeout);
      rt->fw_rc = htonl(krule.fw_rc);
      rt->fw_src_idx = htonl(krule.fw_src_idx);
      rt->fw_src_cnt = htonl(krule.fw_src_cnt);
      rt->fw_dst_idx = htonl(krule.fw_dst_idx);
      rt->fw_dst_cnt = htonl(krule.fw_dst_cnt);
      rt->fw_flags = htonl(krule.fw_flags);
      rt->ttl = htonl((unsigned long)krule.ttl);
      rt->protocol = htonl((unsigned long)krule.protocol);
      rt->fw_rip_idx = htonl(krule.fw_rip_idx);
      rt->fw_rip_cnt = htonl(krule.fw_rip_cnt);
      rt->usage = htonl(krule.usage);
      rt->bytes = htonl(krule.bytes);
      rt->gbytes = htonl(krule.gbytes);
    
      if (write_data(i, lbuf, sizeof(struct serv_data) + sdlength) < 0)
        return;
    }
  } while (result > 0);

  sdlength = 0;
  sd->length = htonl(sdlength);
  write_data(i, lbuf, sizeof(struct serv_data));

  close(fd);

}

/*
 * Send variables to client
 */

static void send_vars(int i)
{ char lbuf[BUFFER_SIZE]; /* use local buffer ! */
  struct serv_data *sd = (struct serv_data *)lbuf;
  struct variable *var;
  struct var_trans {
    long value;
    unsigned long address;
    long timeout;
  };
  struct var_trans *vt = (struct var_trans *)(sd->data);
  int j, vcnt;
  unsigned long sdlength;

  sd->magic = htonl(SERV_MAGIC);
  sd->command = htonl(SERV_FILE_DATA);

  sdlength = 4;
  *((int *)(sd->data)) = htonl((unsigned long)sf_var_free);
  sd->length = htonl(sdlength);
  if (write_data(i, lbuf, sizeof(struct serv_data) + sdlength) < 0) 
    return;

  for (j=0; j<sf_var_free; j++) {
    /* write variable name */
    strcpy(sd->data, varnames[j]);
    sdlength = strlen(varnames[j]);
    sd->length = htonl(sdlength);
    if (write_data(i, lbuf, sizeof(struct serv_data) + sdlength) < 0)
      return;

    /* count entries in variable list including root */
    for (var = &(variables[j]), vcnt = 0; var != NULL; var=var->next, vcnt++);
    sdlength = 4;
    *((int *)(sd->data)) = htonl((unsigned long)vcnt);
    sd->length = htonl(sdlength);
    if (write_data(i, lbuf, sizeof(struct serv_data) + sdlength) < 0) 
      return;
   
    /* send variable list */ 
    sdlength = sizeof(struct var_trans);
    sd->length = htonl(sdlength);
    var = &(variables[j]);
    do {
      vt->value = htonl(var->value);
      vt->address = var->address; /* stored in network byte order */
      vt->timeout = htonl(var->timeout);
      if (write_data(i, lbuf, sizeof(struct serv_data) + sdlength) < 0) 
        return;
      var = var->next;
    } while (var != NULL);

  }
  
  sdlength = 0;
  sd->length = htonl(sdlength);
  write_data(i, lbuf, sizeof(struct serv_data));
}

/*
 * parse config file in child process and
 * return 1 if file is OK, 0 if it contains any errors
 */

static int check_config_file(char *name)
{ int pid;
  int status;
  int i;
  char tmp[256];

  if ((pid = fork()) == 0) { /* child */
    init_temp_data(); /* initialize static variables in sf_parser.y */
    prio_rules = NULL;
    rules = NULL;
    sf_addr_free = 0;
    memset(sf_addr, 0, sizeof(sf_addr));
    for (i = 0; i < SF_ADDRESS_CNT_MAX; i++)
      sf_addr[i].port = USHRT_MAX;
    sf_trusted = 0;
    sf_friends = 0;
    sf_mailaddr = NULL;
    notify = NULL;
    sf_var_free = 0;
    if ((status = sf_read_config_file(name)) != 0) {
      sprintf(tmp, "Error parsing configuration file %s", name);
      log_entry(tmp);
    }
    exit(status);
  }

  waitpid(pid, &status, 0);
  if (!WIFEXITED(status) || (WEXITSTATUS(status) != 0))
    return 0;
  return 1;
}

/*
 * check config file and send result to client
 */

static void check_config(int i) 
{ char lbuf[BUFFER_SIZE]; /* use local buffer ! */
  struct serv_data *sd = (struct serv_data *)lbuf;
  unsigned long sdlength;

  sd->magic = htonl(SERV_MAGIC);

  if (check_config_file(GENERATED_CONFIG)) {
    /* config file ok */
    sd->command = htonl(SERV_OK);
    rename(DEFAULT_CONFIG, DEFAULT_CONFIG".old");
    if (rename(GENERATED_CONFIG, DEFAULT_CONFIG) < 0)
      sd->command = htonl(SERV_ERROR);
  }
  else
    sd->command = htonl(SERV_ERROR);

  sdlength = 0;
  sd->length = htonl(sdlength);
  write_data(i, lbuf, sizeof(struct serv_data));
}

/*
 * Send tcp connections to client
 */

static void send_tcp(int i)
{ char lbuf[BUFFER_SIZE]; /* use local buffer ! */
  struct serv_data *sd = (struct serv_data *)lbuf;
  unsigned long sdlength;
  int fd, result;
  time_t now;
  struct sysinfo info;
  unsigned long *actdate = (unsigned long *)(sd->data);

  sd->magic = htonl(SERV_MAGIC);
  sd->command = htonl(SERV_FILE_DATA);

  now = time(NULL);
  sysinfo(&info);

  if ((fd = open(sf_devname, O_RDONLY)) < 0) {
    sdlength = 0;
    sd->length = htonl(sdlength);
    write_data(i, lbuf, sizeof(struct serv_data));
    return;
  }

  sdlength = sizeof(unsigned long);
  sd->length = htonl(sdlength);
  *actdate = htonl(now);
  if (write_data(i, lbuf, sizeof(struct serv_data) + sdlength) < 0)
    return;

  do {
    struct sf_tcp_connection conn;
    struct tcp_trans {
      unsigned long	id;
      unsigned long     key;
      unsigned long     created;
      unsigned long     created_jiffies;
      unsigned long     lastuse;
      unsigned long     timeout;
      unsigned long     fromaddr;
      unsigned long     fromport;
      unsigned long     toaddr;
      unsigned long     toport;
      unsigned long     allpsource;
      unsigned long     state;
    };
    struct tcp_trans *tt = (struct tcp_trans *)(sd->data);

    sdlength = sizeof(struct tcp_trans);
    sd->length = htonl(sdlength);

    result = read(fd, &conn, sizeof(conn));
    if (result < 0) {
      sdlength = 0;
      sd->length = htonl(sdlength);
      write_data(i, lbuf, sizeof(struct serv_data));
      return;
    }

    if (result > 0) {
      __u32 key;

      if (conn.allpsource) {
        tt->allpsource = htonl(1);
        tt->fromport = htonl(0);
      }
      else {
        tt->allpsource = htonl(0);
        tt->fromport = htonl((unsigned long)ntohs(conn.client_port));
      }

      tt->toport = htonl((unsigned long)ntohs(conn.server_port));

      tt->fromaddr = conn.client_addr;
      tt->toaddr = conn.server_addr;

      tt->created_jiffies = htonl(conn.established);
      tt->created = htonl(now - info.uptime + conn.established / HZ);
      tt->lastuse = htonl(now - info.uptime + conn.lastused / HZ);
      tt->timeout = htonl(conn.timer_set ? 
        now - info.uptime + conn.timer.expires / HZ : 0);

      key = SF_TCP_KEY(conn.client_addr + conn.server_addr +
	    (__u32) conn.client_port + (__u32) conn.server_port);
      tt->key = htonl(key);
      tt->id = htonl(conn.id);
      tt->state = htonl((unsigned long)conn.state);

      if (write_data(i, lbuf, sizeof(struct serv_data) + sdlength) < 0)
        return;
    }
  } while (result > 0);

  sdlength = 0;
  sd->length = htonl(sdlength);
  write_data(i, lbuf, sizeof(struct serv_data));

  close(fd);  
}

/*
 * kill tcp connection(s)
 */

static void kill_tcp(int i, char *data, int size) 
{ struct killdata {
    unsigned long id;
    unsigned long key;
    unsigned long created;
  };
  struct killdata *kd = (struct killdata *)data;
  struct sf_control *control;
  struct sf_control_tcp *ctl;

  switch (size) {
#if 0 /* does not make sense, kills configuration connection, too */
    case 0: /* kill all */

      if ((control = (struct sf_control *) malloc(sizeof(struct sf_control))) == NULL) {
        break;
      }
      control->magic = SF_CONTROL_MAGIC;
      control->command = SF_COMMAND_FLUSH_ALL;

      write(fddev, &control, sizeof(control));

      break;
#endif 
    case sizeof(struct killdata): // kill tcp connection

      if ((control = (struct sf_control *) malloc(sizeof(struct sf_control) +
                     sizeof(struct sf_control_tcp))) == NULL) {
        break;
      }

      control->magic = SF_CONTROL_MAGIC;
      control->command = SF_COMMAND_TCP_KILL;
      ctl = (struct sf_control_tcp *) control->data;

      ctl->id = ntohl(kd->id);
      ctl->key = (unsigned short)ntohl(kd->key);
      ctl->established = ntohl(kd->created);

      write(fddev, control, sizeof(struct sf_control) + sizeof(struct sf_control_tcp));

      free(control);

      break;

    default:
      break;
  }

  /* don't send result, because client connection may have been killed */
}

int seekLogin(struct auth_data *passwdbuf, FILE *passwd, char *login) 
{
	// check for login, key in firewall.passwd
	while (fread(passwdbuf, sizeof(struct auth_data), 1, passwd) == 1) {
		if (passwdbuf->login[strlen(login)] == ' ') {
			passwdbuf->login[strlen(login)] = 0; 			// terminate loginstr
			if (strcmp(passwdbuf->login, login) == 0)
				return 1;
		}
	}
	return 0;
} // seekLogin

/*
 * Read pending data on sockconn and process it
 */

void serv_process(int i)
{ int retval;
  int j;
  struct serv_data *sd;
  unsigned long sdcommand;
  unsigned long sdmagic;
  unsigned long sdlength;
  struct serv_data sdr;
  char tmp[255]; /* string for log message construction */
  char stmp[20]; /* string for IP address to string conversion */

  if ((retval = read_data(i)) <= 0) 
    return;

  /* read successful */
#ifdef SERV_DEBUG
  sprintf(tmp, "read successful, retval == %i", retval);
  debug_entry(i, tmp, 1);
#endif

  restart:

  if ((conn_state[i] & STATE_WAITDATA) != STATE_WAITDATA) {

    if (retval < sizeof(struct serv_data)) { /* wait for more data */
      conn_state[i] |= STATE_WAITDATA;
      if ((connbuf[i] = (char *)malloc(retval)) == NULL)
        crit_error("Cannot allocate memory");
      connbufsize[i] = retval;
      memcpy(connbuf[i], buffer, retval);
      return;
    }

    sd = (struct serv_data *) buffer;
    sdmagic = ntohl(sd->magic);
    sdcommand = ntohl(sd->command);
    sdlength = ntohl(sd->length);

    if ((sdmagic != SERV_MAGIC) || (sdlength > SERV_MAXLEN)) { /* ignore data */
#ifdef SERV_DEBUG
		log_entry("       Wrong magic or length > SERV_MAXLEN");
#endif
      return;
    }

    if ((retval - sizeof(struct serv_data)) < sdlength) { /* wait for more data */
      conn_state[i] |= STATE_WAITDATA;
      if ((connbuf[i] = (char *)malloc(retval)) == NULL)
        crit_error("Cannot allocate memory");
      connbufsize[i] = retval;
      memcpy(connbuf[i], buffer, retval);
      return;
    }
	}
  else { /* assert((conn_state[i] & STATE_WAITDATA) == STATE_WAITDATA) */
    int oldsize = connbufsize[i];

    if ((connbuf[i] = (char *) realloc(connbuf[i], oldsize + retval)) == NULL)
      crit_error("Cannot allocate memory");
    connbufsize[i] += retval;
    memcpy(&(connbuf[i][oldsize]), buffer, retval);
    sd = (struct serv_data *) connbuf[i];
    sdmagic = ntohl(sd->magic);
    sdcommand = ntohl(sd->command);
    sdlength = ntohl(sd->length);

    if ((oldsize + retval) < sizeof(struct serv_data)) { 
			/* wait for more data */
			sprintf(tmp, "Return from handle while size was %i", oldsize + retval);
			log_entry(tmp);
      return;
		}

    if ((sdmagic != SERV_MAGIC) || (sdlength > SERV_MAXLEN)) { /* ignore data */
#ifdef SERV_DEBUG
    log_entry("       Wrong magic or length > SERV_MAXLEN");
#endif
      goto endproc;
    }

    if ((oldsize + retval - sizeof(struct serv_data)) < sdlength) /* wait for more data */
      return;
  }

  /* process data */
#ifdef SERV_DEBUG
  sprintf(tmp, "processing command %lu, length %lu", sdcommand, sdlength);
  debug_entry(i, tmp, 1);
#endif
  if ((conn_state[i] & STATE_WAITHELLO) == STATE_WAITHELLO) {
    if (sdcommand == SERV_CLIENT_HELLO) {
#ifdef SERV_DEBUG
      log_entry("       Received SERV_CLIENT_HELLO");
#endif
      conn_state[i] &= ~STATE_WAITHELLO;
      conn_state[i] |= STATE_NOTAUTHENTICATED;
      goto endproc;
    }
    else {
      strcpy(stmp, inet_ntoa(cliaddr[i].sin_addr));
      sprintf(tmp, "Remote config conn. aborted (no hello): %s:%hu",
              stmp, ntohs(cliaddr[i].sin_port));
      log_entry(tmp);
      handle_abort(i);
      return;
    }
  }
  
  sdr.magic = htonl(SERV_MAGIC);
  sdr.length = 0;

	/* first check if command is reply on our ping */
	if (sdcommand == SERV_PING_REPLY) {
    conn_state[i] &= ~STATE_WAITPING;
#ifdef SERV_DEBUG
      log_entry("       Received SERV_PING_REPLY");
#endif
		goto endproc;
	}

	if ((conn_state[i] & STATE_NOTAUTHENTICATED) == STATE_NOTAUTHENTICATED) {
		if (sdcommand == SERV_AUTHENTICATE) {

			strcpy(key, sd->data);
			strcpy(login, sd->data+MDKEY_SIZE);
			strcpy(stmp, inet_ntoa(cliaddr[i].sin_addr));
			sprintf(tmp, "Authentication: Client %s:%hu",
					stmp, ntohs(cliaddr[i].sin_port));
			log_entry(tmp);
			sprintf(tmp, "Authentication: Login %s", login);
			log_entry(tmp);
			sprintf(tmp, "Authentication: Key ");
			strncat(tmp, key, 32);
			log_entry(tmp);

			MD5Init(&context);
			MD5Update(&context, key, 32);
			MD5Final(digest, &context);
			for (j = 0; j < 16; j++)
				sprintf (&key[j*2], "%02x", digest[j]);

#ifdef SERV_DEBUG
			sprintf(tmp, "New key is: ");
			strncat(tmp, key, 32);
			log_entry(tmp);
#endif

			passwd = fopen(PASSWD, "r");
			if (passwd == 0) {
				sprintf(tmp, "Error opening passwd file: %s", PASSWD);
				log_entry(tmp);
				sdr.command = htonl(SERV_ERROR);
   	   	if (write_data(i, (void *) &sdr, sizeof(sdr)) < 0)
   	     		return;
				goto endproc;
			}

			// check for login, key in firewall.passwd
			if (seekLogin(&passwdbuf, passwd, login) == 1) {
				if (strncmp(passwdbuf.key, key, MDKEY_SIZE) == 0) {
					fclose(passwd);
					passwd = 0;
					conn_state[i] &= ~STATE_NOTAUTHENTICATED;
					// save login for update
					strcpy(&conn_logins[0][i], login);
					// save key for update
					strncpy(&conn_keys[0][i], sd->data, MDKEY_SIZE);
					j = atoi(passwdbuf.keynum); // get keynum
	
					if (j == 1) {
						sprintf(tmp, "Key has expired: Login %s", login);
						log_entry(tmp);
						// send key expired
						sdr.command = htonl(SERV_KEY_EXPIRED);
						if (write_data(i, (void *) &sdr, sizeof(sdr)) < 0)
							return;
						goto endproc;
					}
					else {	
						sprintf(tmp, "Authentication successful: Login %s", login);
						log_entry(tmp);
						sdr.command = htonl(SERV_SERVER_AUTHOK);
						if (write_data(i, (void *) &sdr, sizeof(sdr)) < 0)
							return;
						goto endproc;
					}
				}
			}

			fclose(passwd);
			sprintf(tmp, "Error in Authentication: Login %s", login);
			log_entry(tmp);
			sdr.command = htonl(SERV_ERROR_AUTH);
			if (write_data(i, (void *) &sdr, sizeof(sdr)) < 0)
				return;
	 		goto endproc;		
		}
		else {
      sdr.command = htonl(SERV_NOTAUTHENTICATED);
      if (write_data(i, (void *) &sdr, sizeof(sdr)) < 0)
        return;
			goto endproc;
		}
	}

  switch (sdcommand) {
    case SERV_KEY_UPDATE :
#ifdef SERV_DEBUG
      log_entry("       Received SERV_KEY_UPDATE");
#endif
		passwd = fopen(PASSWD, "r+");
		if (passwd == 0) {
			sprintf(tmp, "Error opening passwd file: %s", PASSWD);
			log_entry(tmp);
			break;
		}
		strcpy(login, &conn_logins[0][i]);
		strncpy(&conn_keys[0][i], sd->data, MDKEY_SIZE);

		// check for login in firewall.passwd

		if (seekLogin(&passwdbuf, passwd, login) == 1) {
			sprintf(tmp, "Updating key for login %s", login);
			log_entry(tmp);
			sprintf(tmp, "New key is %s", &conn_keys[0][i]);
			log_entry(tmp);
			j = 100;		 // default new kwynum 
			sprintf(passwdbuf.keynum, "%4d", j);
			passwdbuf.kspace = ' ';
			strncpy(passwdbuf.key, &conn_keys[0][i], MDKEY_SIZE);
			if(fseek(passwd, -sizeof(passwdbuf), SEEK_CUR) != 0) {
				sprintf(tmp, "Error writing key for login %s", login);
				log_entry(tmp);
				fclose(passwd);
				break;
			}
			if(fwrite(&passwdbuf, MDKEY_SIZE + 5, 1, passwd) != 1) {
				sprintf(tmp, "Error writing key for login %s", login);
				log_entry(tmp);
			}
		}
		fclose(passwd);

		sdr.command = htonl(SERV_OK);
      if (write_data(i, (void *) &sdr, sizeof(sdr)) < 0)
			return;
      break;

    case SERV_CLIENT_AUTHOK :
#ifdef SERV_DEBUG
      log_entry("       Received SERV_CLIENT_AUTHOK");
#endif
		passwd = fopen(PASSWD, "r+");
		if (passwd == 0) {
			sprintf(tmp, "Error opening passwd file: %s", PASSWD);
			log_entry(tmp);
			break;
		}
		strcpy(login, &conn_logins[0][i]);

		// check for login in firewall.passwd
		while (fread(&passwdbuf, sizeof(struct auth_data), 1, passwd) == 1) {
			if (passwdbuf.login[strlen(login)] == ' ') {
				passwdbuf.login[strlen(login)] = 0; 			// terminate loginstr
				if (strcmp(passwdbuf.login, login) == 0) {
//#ifdef SERV_DEBUG
					sprintf(tmp, "Updating key for login %s", login);
					log_entry(tmp);
					sprintf(tmp, "New key is %s", &conn_keys[0][i]);
               log_entry(tmp);
//#endif
					j = atoi(passwdbuf.keynum); // get keynum
					j -= 1;

					sprintf(passwdbuf.keynum, "%4d", j);
					passwdbuf.kspace = ' ';
					strncpy(passwdbuf.key, &conn_keys[0][i], MDKEY_SIZE);
					if(fseek(passwd, -sizeof(passwdbuf), SEEK_CUR) != 0) {
               	sprintf(tmp, "Error writing key for login %s", login);
               	log_entry(tmp);
						break;
					}
					if(fwrite(&passwdbuf, MDKEY_SIZE + 5, 1, passwd) != 1) {
               	sprintf(tmp, "Error writing key for login %s", login);
               	log_entry(tmp);
					}
					break;
				}
			}
      }
		if (passwd != 0)
			fclose(passwd);
      break;

    case SERV_CLOSE:
      sdr.command = htonl(SERV_CLOSE_ACK);
      if (write_data(i, (void *) &sdr, sizeof(sdr)) >= 0) {
        strcpy(stmp, inet_ntoa(cliaddr[i].sin_addr));
        sprintf(tmp, "Remote config conn. closed by client: %s:%hu",
                stmp, ntohs(cliaddr[i].sin_port));
        log_entry(tmp);
        handle_abort(i);
      }
      return;

    case SERV_CLOSE_ACK:
#ifdef SERV_DEBUG
      log_entry("       Received SERV_CLOSE_ACK");
#endif
      if ((conn_state[i] & STATE_CLOSEWAIT) != STATE_CLOSEWAIT)
        break; /* ignore CLOSE_ACK */
      strcpy(stmp, inet_ntoa(cliaddr[i].sin_addr));
      sprintf(tmp, "Remote config conn. closed by server: %s:%hu",
              stmp, ntohs(cliaddr[i].sin_port));
      log_entry(tmp);
      handle_abort(i);
      return;

    case SERV_PING:
#ifdef SERV_DEBUG
      log_entry("       Received SERV_PING");
#endif
      sdr.command = htonl(SERV_PING_REPLY);
      if (write_data(i, (void *) &sdr, sizeof(sdr)) < 0)
        return;
      break;

    case SERV_PING_REPLY :
#ifdef SERV_DEBUG
      log_entry("       Received SERV_PING_REPLY");
#endif
      conn_state[i] &= ~STATE_WAITPING;
      break;

    case SERV_CLIENT_HELLO:
#ifdef SERV_DEBUG
      log_entry("       Received SERV_CLIENT_HELLO");
#endif
      break; /* ignore SERV_CLIENT_HELLO */

    case SERV_GET_FILE:
#ifdef SERV_DEBUG
      sprintf(tmp,"       Received SERV_GET_FILE, %s", sd->data);
      log_entry(tmp);
#endif
      conn_state[i] |= STATE_CHILD;
      if ((conn_child_pid[i] = fork()) == 0) { /* child running */
				sdr.command = htonl(SERV_OK);
				if (write_data(i, (void *) &sdr, sizeof(sdr)) < 0)
					return;
        send_file(i, sd->data);
        exit(0);
      }
      if (conn_child_pid[i] < 0) { /* fork() failed */
        conn_state[i] &= ~STATE_CHILD;
        conn_child_pid[i] = 0;
        log_entry("ERROR: Unable to fork!");
        sdr.command = htonl(SERV_ERROR);
        if (write_data(i, (void *) &sdr, sizeof(sdr)) < 0)
          return;
      }

      break;

    case SERV_WRITE_FILE:
#ifdef SERV_DEBUG
      sprintf(tmp, "Received SERV_WRITE_FILE %s", sd->data);
      log_entry(tmp);
#endif
      strcpy(tmp, CONF_DIR);
      strcat(tmp, sd->data);
      if ((conn_fds[i] = fopen(tmp, "w")) == NULL) {
        sprintf(tmp, "Error creating local file %s%s", CONF_DIR, sd->data);
        log_entry(tmp);
        sdr.command = htonl(SERV_WRITE_FAIL);
        if (write_data(i, (void *) &sdr, sizeof(sdr)) < 0)
          return;
        break;
      }
      conn_state[i] |= STATE_WRITE_FILE;
			sdr.command = htonl(SERV_OK);
			if (write_data(i, (void *) &sdr, sizeof(sdr)) < 0)
        return;
      break;
    
    case SERV_FILE_DATA:
#ifdef SERV_DEBUG
      log_entry("Received SERV_FILE_DATA");
#endif
      if ((conn_state[i] & STATE_WRITE_FILE) != STATE_WRITE_FILE) 
        break;
      if (sdlength == 0) { /* EOF */
        fclose(conn_fds[i]);
        conn_state[i] &= ~STATE_WRITE_FILE;
        sdr.command = htonl(SERV_WRITE_OK);
        if (write_data(i, (void *) &sdr, sizeof(sdr)) < 0)
          return;
        break;
      }
      if (fwrite(sd->data, 1, sdlength, conn_fds[i]) < sdlength) {
        log_entry("Error writing to local file.");
        conn_state[i] &= ~STATE_WRITE_FILE;
        fclose(conn_fds[i]);
        sdr.command = htonl(SERV_WRITE_FAIL);
        if (write_data(i, (void *) &sdr, sizeof(sdr)) < 0)
          return;
      }
			sdr.command = htonl(SERV_OK);
      if (write_data(i, (void *) &sdr, sizeof(sdr)) < 0)
        return;
      break;

    case SERV_READ_ADDR:
#ifdef SERV_DEBUG
      sprintf(tmp,"       Received SERV_READ_ADDR");
      log_entry(tmp);
#endif
      conn_state[i] |= STATE_CHILD;
      if ((conn_child_pid[i] = fork()) == 0) { /* child running */
				sdr.command = htonl(SERV_OK);
				if (write_data(i, (void *) &sdr, sizeof(sdr)) < 0)
					return;
        send_addr(i);
        exit(0);
      }
      if (conn_child_pid[i] < 0) { /* fork() failed */
        conn_state[i] &= ~STATE_CHILD;
        conn_child_pid[i] = 0;
        log_entry("ERROR: Unable to fork!");
        sdr.command = htonl(SERV_FILE_DATA);
        if (write_data(i, (void *) &sdr, sizeof(sdr)) < 0)
          return;
      }

      break;

    case SERV_READ_RULES:
#ifdef SERV_DEBUG
      sprintf(tmp,"       Received SERV_READ_RULES");
      log_entry(tmp);
#endif
      conn_state[i] |= STATE_CHILD;
      if ((conn_child_pid[i] = fork()) == 0) { /* child running */
				sdr.command = htonl(SERV_OK);
				if (write_data(i, (void *) &sdr, sizeof(sdr)) < 0)
					return;
        send_rules(i);
        exit(0);
      }
      if (conn_child_pid[i] < 0) { /* fork() failed */
        conn_state[i] &= ~STATE_CHILD;
        conn_child_pid[i] = 0;
        log_entry("ERROR: Unable to fork!");
        sdr.command = htonl(SERV_FILE_DATA);
        if (write_data(i, (void *) &sdr, sizeof(sdr)) < 0)
          return;
      }

      break;

    case SERV_READ_VARS:
#ifdef SERV_DEBUG
      sprintf(tmp,"       Received SERV_READ_VARS");
      log_entry(tmp);
#endif
      conn_state[i] |= STATE_CHILD;
      if ((conn_child_pid[i] = fork()) == 0) { /* child running */
				sdr.command = htonl(SERV_OK);
				if (write_data(i, (void *) &sdr, sizeof(sdr)) < 0)
					return;
        send_vars(i);
        exit(0);
      }
      if (conn_child_pid[i] < 0) { /* fork() failed */
        conn_state[i] &= ~STATE_CHILD;
        conn_child_pid[i] = 0;
        log_entry("ERROR: Unable to fork!");
        sdr.command = htonl(SERV_FILE_DATA);
        if (write_data(i, (void *) &sdr, sizeof(sdr)) < 0)
          return;
      }

      break;

    case SERV_CHECKCONFIG:
#ifdef SERV_DEBUG
      sprintf(tmp,"       Received SERV_CHECKCONFIG");
      log_entry(tmp);
#endif
      conn_state[i] |= STATE_CHILD;
      if ((conn_child_pid[i] = fork()) == 0) { /* child running */
				sdr.command = htonl(SERV_OK);
				if (write_data(i, (void *) &sdr, sizeof(sdr)) < 0)
					return;
        check_config(i);
        exit(0);
      }
      if (conn_child_pid[i] < 0) { /* fork() failed */
        conn_state[i] &= ~STATE_CHILD;
        conn_child_pid[i] = 0;
        log_entry("ERROR: Unable to fork!");
        sdr.command = htonl(SERV_ERROR);
        if (write_data(i, (void *) &sdr, sizeof(sdr)) < 0)
          return;
      }

      break;

    case SERV_RECONFIG:
#ifdef SERV_DEBUG
      sprintf(tmp,"       Received SERV_RECONFIG");
      log_entry(tmp);
#endif
      if (!check_config_file(current_config)) {
        sdr.command = htonl(SERV_ERROR);
        if (write_data(i, (void *) &sdr, sizeof(sdr)) < 0)
          return;
      }
      else {
        daemon_resetusage();
        sdr.command = htonl(SERV_OK);
        if (write_data(i, (void *) &sdr, sizeof(sdr)) < 0)
          return;
        abort();
      }

      break;

    case SERV_READ_TCP:
#ifdef SERV_DEBUG
      sprintf(tmp,"       Received SERV_READ_TCP");
      log_entry(tmp);
#endif
      conn_state[i] |= STATE_CHILD;
      if ((conn_child_pid[i] = fork()) == 0) { /* child running */
				sdr.command = htonl(SERV_OK);
				if (write_data(i, (void *) &sdr, sizeof(sdr)) < 0)
					return;
        send_tcp(i);
        exit(0);
      }
      if (conn_child_pid[i] < 0) { /* fork() failed */
				sdr.command = htonl(SERV_ERROR);
				if (write_data(i, (void *) &sdr, sizeof(sdr)) < 0)
					return;
        conn_state[i] &= ~STATE_CHILD;
        conn_child_pid[i] = 0;
        log_entry("ERROR: Unable to fork!");
        sdr.command = htonl(SERV_FILE_DATA);
        if (write_data(i, (void *) &sdr, sizeof(sdr)) < 0)
          return;
      }

      break;

    case SERV_KILL_TCP:
#ifdef SERV_DEBUG
      sprintf(tmp,"       Received SERV_KILL_TCP");
      log_entry(tmp);
#endif
      conn_state[i] |= STATE_CHILD;
      if ((conn_child_pid[i] = fork()) == 0) { /* child running */
				sdr.command = htonl(SERV_OK);
				if (write_data(i, (void *) &sdr, sizeof(sdr)) < 0)
					return;
        kill_tcp(i, sd->data, sdlength);
        exit(0);
      }
      if (conn_child_pid[i] < 0) { /* fork() failed */
        conn_state[i] &= ~STATE_CHILD;
        conn_child_pid[i] = 0;
        log_entry("ERROR: Unable to fork!");
        sdr.command = htonl(SERV_ERROR);
        if (write_data(i, (void *) &sdr, sizeof(sdr)) < 0)
          return;
      }

      break;

    case SERV_RESET_USAGE:
#ifdef SERV_DEBUG
      sprintf(tmp,"       Received SERV_RESET_USAGE");
      log_entry(tmp);
#endif
      daemon_resetusage();
			sdr.command = htonl(SERV_OK);
			if (write_data(i, (void *) &sdr, sizeof(sdr)) < 0)
				return;
      break;

    default:
#ifdef SERV_DEBUG
      sprintf(tmp, "Received unknown command: %lu", sdcommand);
      log_entry(tmp);
#endif
      break;
  }

  endproc:

  /* free memory */
  if ((conn_state[i] & STATE_WAITDATA) == STATE_WAITDATA) {
    if ((connbufsize[i] - sizeof(struct serv_data)) > sdlength) {

      /* beginning of next command may be in this packet */
      retval = connbufsize[i] - sizeof(struct serv_data) - sdlength;
      memcpy(buffer, &(connbuf[i][sizeof(struct serv_data) + sdlength]), 
             retval);
      free(connbuf[i]);
      connbufsize[i] = 0;
      conn_state[i] &= ~STATE_WAITDATA;
#ifdef SERV_DEBUG
      log_entry("       Data left, restarting function");
#endif
      goto restart;
    }
    free(connbuf[i]);
    connbufsize[i] = 0;
    conn_state[i] &= ~STATE_WAITDATA;
  }
  else 
    if ((retval - sizeof(struct serv_data)) > sdlength) {
      /* beginning of next command may be in this packet */
      retval = retval - sizeof(struct serv_data) - sdlength;
      memmove(buffer, &(buffer[sizeof(struct serv_data) + sdlength]), retval);
#ifdef SERV_DEBUG
      log_entry("       Data left, restarting function");
#endif
      goto restart;
    }

  return;

}
